/*
* Copyright (C) 2011 Sunil Mohan Adapa <sunil@medhas.org>.
* Copyright (C) 2011 O S K Chaitanya <oskchaitanya@gmail.com>.
*
* Author: Sunil Mohan Adapa <sunil@medhas.org>
*         O S K Chaitanya <oskchaitanya@gmail.com>
*
* This file is part of GNOME Nonogram.
*
* GNOME Nonogram is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* GNOME Nonogram is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with GNOME Nonogram. If not, see <http://www.gnu.org/licenses/>.
*/

const Lang = imports.lang;
const GObject = imports.gi.GObject;
const Gtk = imports.gi.Gtk;

const Main = imports.main; // FIXME:
const NonogramIO = imports.nonogramio;

GameWidget = new GType(_GameWidget = {
    parent: Gtk.Table.type,
    name: "GameWidget",
    properties: [
        {
            name: "puzzleFile",
            type: GObject.TYPE_STRING,
            default_value: null,
            flags: GObject.ParamFlags.READABLE
                 | GObject.ParamFlags.WRITABLE
                 | GObject.ParamFlags.CONSTRUCT
        }
    ],

    _BACKGROUND_COLOR: 0xffffffff,
    _NO_COLOR: -1,

    _incorrectDialogShown: false,
    _rowCluesTable: null,
    _columnCluesTable: null,
    _pixelGrid: null,

    class_init: function(klass, prototype) {
        prototype._BACKGROUND_COLOR = _GameWidget._BACKGROUND_COLOR;
        prototype._NO_COLOR = _GameWidget._NO_COLOR;

        prototype._incorrectDialogShown = _GameWidget._incorrectDialogShown;
        prototype._rowCluesTable = _GameWidget._rowCluesTable;
        prototype._columnCluesTable = _GameWidget._columnCluesTable;
        prototype._pixelGrid = _GameWidget._pixelGrid;

        prototype.__defineGetter__("_puzzleFile", _GameWidget._getPuzzleFile);
        prototype.__defineSetter__("_puzzleFile", _GameWidget._setPuzzleFile);
        prototype.__defineGetter__("puzzleFile", _GameWidget._getPuzzleFile);
        prototype.__defineSetter__("puzzleFile", _GameWidget._setPuzzleFile);

        prototype._onRealize = _GameWidget._onRealize;
        prototype._listColorIndices = _GameWidget._listColorIndices;
        prototype._generateClues = _GameWidget._generateClues;
        prototype._generateRowClues = _GameWidget._generateRowClues;
        prototype._generateColumnClues = _GameWidget._generateColumnClues;
        prototype._updateClueTables = _GameWidget._updateClueTables;
        prototype._updatePixelGrid = _GameWidget._updatePixelGrid;
        prototype._createBlankPuzzlePixels =
            _GameWidget._createBlankPuzzlePixels;
        prototype._onPuzzleChanged = _GameWidget._onPuzzleChanged;
        prototype._onPuzzleFull = _GameWidget._onPuzzleFull;
        prototype._isPuzzleSolved = _GameWidget._isPuzzleSolved;
        prototype._onPuzzleSolved = _GameWidget._onPuzzleSolved;
        prototype._onPuzzleIncorrectSolution =
            _GameWidget._onPuzzleIncorrectSolution;
    },

    init: function() {
        this.signal.realize.connect(Lang.bind(this, this._onRealize));
    },

    _onRealize: function() {
        this._rowCluesTable = Main.builder.get_object("row_clues_table");
        this._columnCluesTable = Main.builder.get_object("column_clues_table");
        this._pixelGrid = Main.builder.get_object("pixelgrid");

        this._pixelGrid.signal.puzzle_changed.connect(
            Lang.bind(this, this._onPuzzleChanged));
        this._pixelGrid.signal.puzzle_full.connect(
            Lang.bind(this, this._onPuzzleFull));
    },

    _getPuzzleFile: function() {
        return this.__puzzleFile;
    },

    _setPuzzleFile: function(value) {
        try {
            this._solutionPuzzlePixels = NonogramIO.load_puzzle(value);
        } catch (e) {
            if (e.name == "PuzzleLoadFileError")
                print("File error while loading puzzle: " + e.message);
            else if (e.name == "PuzzleLoadImageError")
                print("Image error while loading puzzle: " + e.message);
            return;
        }

        this._listColorIndices();
        this._generateClues();

        this._updateClueTables();
        this._updatePixelGrid();

        this.__puzzleFile = value;
        this._incorrectDialogShown = false;
    },

    _listColorIndices: function() {
        this._colorIndices = {};
        for (var y = 0; y < this._solutionPuzzlePixels.length; ++y) {
            for (var x = 0; x < this._solutionPuzzlePixels[y].length; ++x) {
                this._colorIndices[this._solutionPuzzlePixels[y][x]] = 1;
            }
        }
    },

    _generateClues: function() {
        this._generateRowClues();
        this._generateColumnClues();
    },

    _generateRowClues: function() {
        this._rowClues = [];
        for (var y = 0; y < this._solutionPuzzlePixels.length; ++y) {
            var singleRowClue = [];
            var currentColor = this._BACKGROUND_COLOR;
            var currentCount = 0;
            for (var x = 0; x < this._solutionPuzzlePixels[y].length; ++x) {
                if (this._solutionPuzzlePixels[y][x] == currentColor)
                    currentCount++;
                else {
                    if (currentColor != this._BACKGROUND_COLOR)
                        singleRowClue.push({color: currentColor,
                                            count: currentCount});

                    currentColor = this._solutionPuzzlePixels[y][x];
                    currentCount = 1;
                }
            }
            if (currentColor != this._BACKGROUND_COLOR)
                singleRowClue.push({color: currentColor,
                                    count: currentCount});

            this._rowClues.push(singleRowClue);
        }
    },

    _generateColumnClues: function() {
        this._columnClues = [];
        for (var x = 0; x < this._solutionPuzzlePixels[0].length; ++x) {
            var singleColumnClue = [];
            var currentColor = this._BACKGROUND_COLOR;
            var currentCount = 0;
            for (var y = 0; y < this._solutionPuzzlePixels.length; ++y) {
                if (this._solutionPuzzlePixels[y][x] == currentColor)
                    currentCount++;
                else {
                    if (currentColor != this._BACKGROUND_COLOR)
                        singleColumnClue.push({color: currentColor,
                                               count: currentCount});

                    currentColor = this._solutionPuzzlePixels[y][x];
                    currentCount = 1;
                }
            }
            if (currentColor != this._BACKGROUND_COLOR)
                singleColumnClue.push({color: currentColor,
                                       count: currentCount});

            this._columnClues.push(singleColumnClue);
        }
    },

    _updateClueTables: function() {
        this._rowCluesTable.setData(this._rowClues);
        this._columnCluesTable.setData(this._columnClues);
    },

    _updatePixelGrid: function() {
        this._pixelGrid.puzzlePixels = this._createBlankPuzzlePixels();
        for(color in this._colorIndices) {
            if(color != this._BACKGROUND_COLOR) {
                this._pixelGrid.drawingColor = color;
                break;
            }
        }
    },

    _createBlankPuzzlePixels: function() {
        var blankPuzzlePixels = [];
        for(var y = 0; y < this._solutionPuzzlePixels.length; ++y) {
            blankPuzzlePixels[y] = [];
            for(var x = 0; x < this._solutionPuzzlePixels[0].length; ++x) {
                blankPuzzlePixels[y][x] = this._NO_COLOR;
            }
        }

        return blankPuzzlePixels;
    },

    _onPuzzleChanged: function() {
        var isPuzzleSolved = this._isPuzzleSolved();

        if(isPuzzleSolved)
            this._onPuzzleSolved();
    },

    _onPuzzleFull: function() {
        if(this._isPuzzleSolved() == false)
            this._onPuzzleIncorrectSolution();
    },

    _isPuzzleSolved: function() {
        var puzzlePixels = this._pixelGrid.puzzlePixels;

        for(var y = 0; y < this._solutionPuzzlePixels.length; ++y) {
            for(var x = 0; x < this._solutionPuzzlePixels[0].length; ++x) {
                if(this._solutionPuzzlePixels[y][x] == this._BACKGROUND_COLOR
                    && puzzlePixels[y][x] == this._NO_COLOR)
                    continue;

                if(this._solutionPuzzlePixels[y][x] != puzzlePixels[y][x])
                    return false;
            }
        }

        return true;
    },

    _onPuzzleSolved: function() {
        dialog = Main.builder.get_object("correct_solution_dialog");
        dialog.run();
        dialog.hide();
        // TODO: disable puzzle interface
    },

    _onPuzzleIncorrectSolution: function() {
        if(this._incorrectDialogShown)
            return;

        this._incorrectDialogShown = true;

        dialog = Main.builder.get_object("incorrect_solution_dialog");
        dialog.run();
        dialog.hide();
    }
});
