#include <glib.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <seed.h>

#define SUPPORTED_BITS_PER_SAMPLE 8
#define SUPPORTED_COLORSPACE GDK_COLORSPACE_RGB

static GdkPixbuf * load_puzzle_file(guchar *path,
                                    SeedContext ctx,
                                    SeedException *exception);
static SeedValue puzzle_new_from_pixbuf(const GdkPixbuf *pixbuf,
                                        SeedContext ctx,
                                        SeedException *exception);

SeedValue
load_puzzle(SeedContext ctx,
            SeedObject function,
            SeedObject this_object,
            gsize argument_count,
            const SeedValue arguments[],
            SeedException *exception)
{
  guchar *path;
  GdkPixbuf *pixbuf;
  SeedValue puzzle;

  // FIXME: Use CHECK_ARG_COUNT macro instead.
  // Currently, including seed-module.h results in build failure because
  // of mismatched include path and location of header file.
  if (argument_count != 1)
    {
      seed_make_exception(ctx, exception, "ArgumentError",
                          "wrong number of arguments; expected %u, got %lu",
                          1, (unsigned long) argument_count);
      return seed_make_undefined(ctx);
    }

  path = seed_value_to_string(ctx, arguments[0], exception);

  pixbuf = load_puzzle_file(path, ctx, exception);
  if (pixbuf == NULL)
    return seed_make_undefined(ctx);

  puzzle = puzzle_new_from_pixbuf(pixbuf, ctx, exception);
  if (puzzle == NULL)
    return seed_make_undefined(ctx);

  g_free(path);

  return puzzle;
}

static GdkPixbuf *
load_puzzle_file(guchar * path,
                 SeedContext ctx,
                 SeedException *exception)
{
  GError *error = NULL;
  GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(path, &error);
  if (pixbuf == NULL)
    {
      char *error_type = "PuzzleLoadFileError";
      if (error->domain == GDK_PIXBUF_ERROR)
        error_type = "PuzzleLoadImageError";

      seed_make_exception(ctx, exception, error_type, "%s", error->message);
      g_error_free(error);
      return NULL;
    }

  return pixbuf;
}

static SeedValue
puzzle_new_from_pixbuf(const GdkPixbuf *pixbuf,
                       SeedContext ctx,
                       SeedException *exception) {
  int bits_per_sample = gdk_pixbuf_get_bits_per_sample(pixbuf);
  if (bits_per_sample != SUPPORTED_BITS_PER_SAMPLE)
    {
      seed_make_exception(ctx, exception, "PuzzleLoadImageError",
                          "Image has %d bits per sample, %d expected",
                          bits_per_sample, SUPPORTED_BITS_PER_SAMPLE);
      return NULL;
    }

  int num_channels = gdk_pixbuf_get_n_channels(pixbuf);
  if (num_channels != 3 && num_channels != 4)
    {
      seed_make_exception(ctx, exception, "PuzzleLoadImageError",
                          "Image has %d channels, 3 or 4 channels expected",
                          num_channels);
      return NULL;
    }

  GdkColorspace colorspace = gdk_pixbuf_get_colorspace(pixbuf);
  if (colorspace != SUPPORTED_COLORSPACE)
    {
      seed_make_exception(ctx, exception, "PuzzleLoadImageError",
                          "Image colorspace not supported, expected RGB");
      return NULL;
    }

  guchar *pixels = gdk_pixbuf_get_pixels(pixbuf);
  guint width = gdk_pixbuf_get_width(pixbuf);
  guint height = gdk_pixbuf_get_height(pixbuf);
  guint rowstride = gdk_pixbuf_get_rowstride(pixbuf);

  SeedValue puzzle_seed_array;
  SeedValue puzzle[height];
  int x, y;
  for (y = 0; y < height; ++y)
    {
      SeedValue row[width];
      for (x = 0; x < width; ++x)
        {
          guint32 color;
          guint32 r = pixels[y * rowstride + x * num_channels + 0];
          guint32 g = pixels[y * rowstride + x * num_channels + 1];
          guint32 b = pixels[y * rowstride + x * num_channels + 2];
          guint32 a = 255;
          if (num_channels == 4)
            a = pixels[y * rowstride + x * num_channels + 3];

          color = (r << 24) + (g << 16) + (b << 8) + a;

          row[x] = seed_value_from_uint(ctx, color, exception);
          if (row[x] == NULL)
            return NULL;
        }
      puzzle[y] = seed_make_array(ctx, row, width, exception);
      if (puzzle[y] == NULL)
        return NULL;
    }

  puzzle_seed_array = seed_make_array(ctx, puzzle, height, exception);
  return puzzle_seed_array;
}

seed_static_function static_functions[] = {
  {"load_puzzle", load_puzzle, 0}
};

SeedObject
seed_module_init(SeedEngine *engine)
{
  SeedObject nonogram_io;
  SeedGlobalContext *ctx = engine->context;

  seed_class_definition class_def = seed_empty_class;
  class_def.static_functions = static_functions;

  SeedClass nonogram_io_class = seed_create_class(&class_def);

  nonogram_io = seed_make_object(ctx, nonogram_io_class, NULL);
  seed_value_protect(ctx, nonogram_io);

  return nonogram_io;
}
